from typing import Union

from fastapi import FastAPI
from fastapi.exceptions import RequestValidationError
from pydantic import ValidationError
from starlette import status
from najapy.common.async_base import Utils
from najapy.frame.fastapi.base import Request

from najapy.scaffold.fastapi_example_project.fastapi_example.utils.response import RespCode, ErrorResponse


def exception_handler(fastapi: FastAPI):
    fastapi.add_exception_handler(RequestValidationError, request_validation_exception_handler)
    fastapi.add_exception_handler(ValidationError, request_validation_exception_handler)


async def request_validation_exception_handler(
    _: Request,
    exc: Union[RequestValidationError, ValidationError]
) -> ErrorResponse:
    Utils.log.debug(f"[request_validation_exception_handler] {exc.errors()=}")

    return ErrorResponse(
        error_code=RespCode.InvalidParams,
        status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
        details=exc.errors()
    )



from najapy.common.struct import Const, Result
from starlette.responses import UJSONResponse

from najapy.frame.fastapi.response import ErrorResponse as _ErrorResponse


class RespCode():
    Success = 0  # 成功
    Unknown = -1  # 未知
    Database = -2  # 数据库异常
    LackParams = -3  # 缺少必要请求参数
    DataUnExcept = -4  # 异常数据，暂无法操作
    NetError = -5  # 网络错误
    TokenInvalid = -6  # token 无效
    NoPermission = -7  # 无权操作
    InvalidParams = -8  # 无效的参数
    NotSupport = -9  # 不支持的业务类型


RespMessage = Const()
RespMessage[RespCode.Success] = r"成功"
RespMessage[RespCode.Unknown] = r"未知"
RespMessage[RespCode.Database] = r"数据库异常"
RespMessage[RespCode.LackParams] = r"缺少必要请求参数"
RespMessage[RespCode.DataUnExcept] = r'异常数据，暂无法操作'
RespMessage[RespCode.NetError] = r'网络错误，暂无法操作'
RespMessage[RespCode.TokenInvalid] = r"token 无效"
RespMessage[RespCode.NoPermission] = r"无权操作"
RespMessage[RespCode.InvalidParams] = r'无效的参数'
RespMessage[RespCode.NotSupport] = r'不支持的业务类型'


class Response(UJSONResponse):

    def render(self, content):
        return super().render(Result(code=RespCode.Success, data=content, msg=RespMessage[RespCode.Success]))


class ErrorResponse(_ErrorResponse):

    def __init__(self, error_code, content=None, status_code=200, details=None, **kwargs):
        super().__init__(
            error_code=error_code, content=content, msg=RespMessage[error_code], status_code=status_code,
            details=details, **kwargs
        )



# -*- coding: utf-8 -*-
from typing import List, Any, Dict, Union

from najapy.common.async_base import Utils
from pydantic.fields import ModelField


class _BaseForm:

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v: Any, field: ModelField) -> Any:
        raise NotImplementedError

    @classmethod
    def validate_value_type(cls, v, ele_type):
        if not v:
            return v

        try:
            v = ele_type(v)
        except Exception as _:
            raise ValueError(f'Invalid field value:{v} trans type {ele_type=}')

        return v

    @classmethod
    def validate_list_value(cls, v, ele_type=str, sep=','):
        """校验列表元素字段值"""
        if not v:
            return []

        try:

            return [ele_type(x.strip()) for x in v.split(sep)]

        except Exception as _:
            raise ValueError(f'Invalid list field value:{v}')

    @classmethod
    def validate_json_value(cls, v):
        """校验json字段值"""
        if not v:
            return None

        try:

            return Utils.json_decode(v)

        except Exception as _:
            raise ValueError(f'Invalid json field value:{v}')

    @classmethod
    def validate_length_value(cls, v, length=2):
        """校验字段值长度为限定值"""
        if not v:
            return []

        if len(v) != length:
            raise ValueError(f'Invalid list field value:{v}')

        return v

    def __repr__(self):
        return f"{self.__class__.__name__}({super().__repr__()})"


class _BaseEnumForm(_BaseForm):
    _TYPE = None

    @classmethod
    def validate(cls, v: Any, field: ModelField) -> Union[str, int]:
        validate_v = cls.validate_value_type(v, cls._TYPE)

        if scope := field.field_info.extra.get('status_in', []):

            if validate_v not in scope:
                raise ValueError(f'Invalid field value:{v}')

        return validate_v


class _BaseListForm(_BaseForm):
    _TYPE = None
    _LENGTH = None

    @classmethod
    def validate(cls, v: str, field: ModelField) -> List:

        validate_v = cls.validate_list_value(v, ele_type=cls._TYPE)

        if scope := field.field_info.extra.get('status_in', []):

            for _temp in validate_v:

                if _temp not in scope:
                    raise ValueError(f'Invalid list field value:{v}')

        if cls._LENGTH:
            cls.validate_length_value(validate_v, length=cls._LENGTH)

        return validate_v


class IntEnumForm(int, _BaseEnumForm):
    _TYPE = int


class StrEnumForm(str, _BaseEnumForm):
    _TYPE = str


class IntListForm(str, _BaseListForm):
    _TYPE = int


class IntListLength2Form(IntListForm):
    _LENGTH = 2


class StrListForm(str, _BaseForm):
    _TYPE = str


class StrListLength2Form(StrListForm):
    _LENGTH = 2


class JsonForm(str, _BaseForm):
    @classmethod
    def validate(cls, v: str, field: ModelField) -> Dict:
        validate_res = cls.validate_json_value(v)

        return validate_res



# -*- coding: utf-8 -*-

import os

from najapy.common.struct import Configure

BASE_PATH = os.path.abspath(os.path.dirname(__file__))


class _Static(Configure):

    def _init_options(self):
        ##################################################
        # MySql数据库

        self.MySqlMasterServer = self._parser.get_split_host(r'MySql', r'MySqlMasterServer')

        self.MySqlSlaveServer = self._parser.get_split_host(r'MySql', r'MySqlSlaveServer')

        self.MySqlName = self._parser.get(r'MySql', r'MySqlName')

        self.MySqlUser = self._parser.get(r'MySql', r'MySqlUser')

        self.MySqlPasswd = self._parser.get(r'MySql', r'MySqlPasswd')

        self.MySqlMasterMinConn = self._parser.getint(r'MySql', r'MySqlMasterMinConn')

        self.MySqlMasterMaxConn = self._parser.getint(r'MySql', r'MySqlMasterMaxConn')

        self.MySqlSlaveMinConn = self._parser.getint(r'MySql', r'MySqlSlaveMinConn')

        self.MySqlSlaveMaxConn = self._parser.getint(r'MySql', r'MySqlSlaveMaxConn')

        ##################################################
        # Mongo数据库
        self.MongoServer = self._parser.get_split_str('Mongo', 'MongoHost')
        self.MongoName = self._parser.get('Mongo', 'MongoName')
        self.MongoUser = self._parser.get('Mongo', 'MongoUser', fallback=None)
        self.MongoPasswd = self._parser.get('Mongo', 'MongoPasswd', fallback=None)
        self.MongoMinConn = self._parser.getint('Mongo', 'MongoMinConn')
        self.MongoMaxConn = self._parser.getint('Mongo', 'MongoMaxConn')

        ##################################################
        # 缓存

        self.RedisHost = self._parser.get_split_host(r'Redis', r'RedisHost')

        self.RedisBase = self._parser.getint(r'Redis', r'RedisBase')

        self.RedisPasswd = self._parser.get(r'Redis', r'RedisPasswd')

        self.RedisMinConn = self._parser.getint(r'Redis', r'RedisMinConn')

        self.RedisMaxConn = self._parser.getint(r'Redis', r'RedisMaxConn')

        self.RedisExpire = self._parser.getint(r'Redis', r'RedisExpire')

        self.RedisKeyPrefix = self._parser.get(r'Redis', r'RedisKeyPrefix')


class _Dynamic(Configure):

    def _init_options(self):
        ##################################################
        # 基本

        self.Port = self._parser.getint(r'Base', r'Port')
        self.Debug = self._parser.getboolean(r'Base', r'Debug')
        allowed_hosts = self._parser.get_split_str(r'Base', r'AllowedHosts')
        self.AllowedHosts = list(allowed_hosts) or []
        abnormal_status = self._parser.get_split_int(r'Base', r'AbnormalStatus')
        self.AbnormalStatus = list(abnormal_status) or []
        self.JWTSecret = self._parser.get(r"Base", r"JWTSecret")

        ##################################################
        # 日志

        self.LogLevel = self._parser.get(r'Log', r'LogLevel')
        self.LogFilePath = self._parser.get(r'Log', r'LogFilePath')

        ##################################################
        # 事件总线

        self.ClentEventBusChannels = self._parser.getint(r'Event', r'ClentEventBusChannels')


ConfigStatic = _Static()
ConfigDynamic = _Dynamic()

cluster = os.getenv(r'CLUSTER', None)

if cluster is None:
    ConfigStatic.read(f'{BASE_PATH}/static.conf')
    ConfigDynamic.read(f'{BASE_PATH}/dynamic.conf')
else:
    ConfigStatic.read(f'{BASE_PATH}/static.{cluster.lower()}.conf')
    ConfigDynamic.read(f'{BASE_PATH}/dynamic.{cluster.lower()}.conf')



import asyncio

from cachetools import TTLCache

from najapy.common.async_base import Utils


class StackCache:
    """堆栈缓存

    使用运行内存作为高速缓存，可有效提高并发的处理能力

    """

    def __init__(self, maxsize=0xff, ttl=60):

        self._cache = TTLCache(maxsize, ttl)

    def has(self, key):

        return key in self._cache

    def get(self, key, default=None):

        return self._cache.get(key, default)

    def set(self, key, val):

        self._cache[key] = val

    def incr(self, key, val=1):

        res = self.get(key, 0) + val

        self.set(key, res)

        return res

    def decr(self, key, val=1):

        res = self.get(key, 0) - val

        self.set(key, res)

        return res

    def delete(self, key):

        del self._cache[key]

    def size(self):

        return len(self._cache)

    def clear(self):

        return self._cache.clear()


class FuncCache:
    """函数缓存

    使用堆栈缓存实现的函数缓存，在有效期内函数签名一致就会命中缓存

    """

    def __init__(self, maxsize=0xff, ttl=10):

        self._cache = StackCache(maxsize, ttl)

    def __call__(self, func):

        @Utils.func_wraps(func)
        async def _wrapper(*args, **kwargs):

            func_sign = Utils.params_sign(func, *args, **kwargs)

            result = self._cache.get(func_sign)

            if result is None:

                result = await Utils.awaitable_wrapper(
                    func(*args, **kwargs)
                )

                if result is not None:
                    self._cache.set(func_sign, result)

            return result

        return _wrapper


class ShareFuture:
    """共享Future装饰器

    同一时刻并发调用函数时，使用该装饰器的函数签名一致的调用，会共享计算结果

    """

    def __init__(self):

        self._future = {}

    def __call__(self, func):

        @Utils.func_wraps(func)
        async def _wrapper(*args, **kwargs):

            future = None

            func_sign = Utils.params_sign(func, *args, **kwargs)

            if func_sign in self._future:

                future = asyncio.Future()

                self._future[func_sign].append(future)

            else:

                future = Utils.create_task(
                    func(*args, **kwargs)
                )

                if future is None:
                    TypeError(r'Not Coroutine Object')

                self._future[func_sign] = [future]

                future.add_done_callback(
                    Utils.func_partial(self._clear_future, func_sign)
                )

            return await future

        return _wrapper

    def _clear_future(self, func_sign, _):

        if func_sign not in self._future:
            return

        futures = self._future.pop(func_sign)

        result = futures.pop(0).result()

        for future in futures:
            future.set_result(Utils.deepcopy(result))



from motor.motor_asyncio import AsyncIOMotorClient

from najapy.common.async_base import Utils

MONGO_POLL_WATER_LEVEL_WARNING_LINE = 0x08


class MongoPool:
    """Mongo连接管理
    """

    def __init__(
            self, address=None, host=None, username=None, password=None,
            *, name=None, min_pool_size=8, max_pool_size=32, max_idle_time=3600, wait_queue_timeout=10,
            compressors=r'zlib', zlib_compression_level=6,
            **settings):

        self._name = name if name is not None else Utils.uuid1()[:8]

        settings[r'minPoolSize'] = min_pool_size
        settings[r'maxPoolSize'] = max_pool_size
        settings[r'maxIdleTimeMS'] = max_idle_time * 1000
        settings[r'waitQueueTimeoutMS'] = wait_queue_timeout * 1000
        settings[r'compressors'] = compressors
        settings[r'zlibCompressionLevel'] = zlib_compression_level

        if address:
            self._pool = AsyncIOMotorClient(address, **settings)

        else:
            settings[r'host'] = host
            settings[r'username'] = username
            settings[r'password'] = password

            self._pool = AsyncIOMotorClient(**settings)

        for server in self._servers.values():
            server.pool.remove_stale_sockets()

        Utils.log.info(
            f"Mongo {address=} {host=} ({self._name}) "
            f"initialized: {self._pool.min_pool_size}/{self._pool.max_pool_size}"
            f"{settings}"
        )

    @property
    def _servers(self):

        return self._pool.delegate._topology._servers

    def _echo_pool_info(self):

        global MONGO_POLL_WATER_LEVEL_WARNING_LINE

        for address, server in self._servers.items():

            poll_size = len(server.pool.sockets) + server.pool.active_sockets

            if (self._pool.max_pool_size - poll_size) < MONGO_POLL_WATER_LEVEL_WARNING_LINE:
                Utils.log.warning(
                    f'Mongo connection pool not enough ({self._name}){address}: '
                    f'{poll_size}/{self._pool.max_pool_size}'
                )
            else:
                Utils.log.debug(
                    f'Mongo connection pool info ({self._name}){address}: '
                    f'{poll_size}/{self._pool.max_pool_size}'
                )

    def reset(self):

        for address, server in self._servers.items():
            server.pool.reset()
            server.pool.remove_stale_sockets()

            Utils.log.info(
                f'Mongo connection pool reset {address}: {len(server.pool.sockets)}/{self._pool.max_pool_size}'
            )

    def close(self):

        if self._pool is not None:
            self._pool.close()
            self._pool = None

    def get_database(self, db_name):

        self._echo_pool_info()

        result = None

        try:
            result = self._pool[db_name]
        except Exception as err:
            Utils.log.exception(err)

        return result


class MongoDelegate:
    """Mongo功能组件
    """

    def __init__(self, *args, **kwargs):

        self._mongo_pool = MongoPool(*args, **kwargs)

    @property
    def mongo_pool(self):

        return self._mongo_pool

    async def mongo_health(self):

        result = False

        try:
            result = bool(await self._mongo_pool._pool.server_info())
        except Exception as err:
            Utils.log.error(err)

        return result

    def reset_mongo_pool(self):

        self._mongo_pool.reset()

    def close_mongo_pool(self):

        self._mongo_pool.close()

    def get_mongo_database(self, db_name):

        return self._mongo_pool.get_database(db_name)

    def get_mongo_collection(self, db_name, collection):

        return self.get_mongo_database(db_name)[collection]



import aioredis
from aioredis.errors import ReplyError, MaxClientsError, AuthError, ReadOnlyError
from aioredis.util import _NOTSET
from aioredis.commands.string import StringCommandsMixin
from aioredis.commands.transaction import Pipeline, MultiExec

from najapy.common.async_base import Utils, AsyncContextManager, AsyncCirculator
from najapy.common.base import WeakContextVar
from najapy.event.async_event import DistributedEvent

REDIS_ERROR_RETRY_COUNT = 0x1f
REDIS_POOL_WATER_LEVEL_WARNING_LINE = 0x08


class RedisPool:
    """Redis连接管理
    """

    def __init__(self, address, password=None, *, minsize=8, maxsize=32, db=0, expire=0, key_prefix=None, **settings):

        self._pool = None
        self._expire = expire
        self._key_prefix = key_prefix

        self._settings = settings

        self._settings[r'address'] = address
        self._settings[r'password'] = password

        self._settings[r'minsize'] = minsize
        self._settings[r'maxsize'] = maxsize

        self._settings[r'db'] = db

    def __await__(self):

        self._pool = yield from aioredis.create_pool(**self._settings).__await__()

        Utils.log.info(f"Redis {self._settings[r'address']} initialized: {self._pool.size}/{self._pool.maxsize}")

        return self

    async def close(self):

        if self._pool is not None:
            self._pool.close()
            await self._pool.wait_closed()

            self._pool = None

    def get_client(self):

        client = None

        if self._pool is not None:
            client = CacheClient(self._pool, self._expire, self._key_prefix)

        return client


class RedisDelegate:
    """Redis功能组件
    """

    def __init__(self):

        self._redis_pool = None

        self._cache_client_context = WeakContextVar(f'cache_client_{Utils.uuid1()}')

    @property
    def redis_pool(self):

        return self._redis_pool

    async def async_init_redis(self, *args, **kwargs):

        self._redis_pool = await RedisPool(*args, **kwargs)

    async def async_close_redis(self):

        if self._redis_pool is not None:
            await self._redis_pool.close()
            self._redis_pool = None

    async def cache_health(self):

        result = False

        async with self.get_cache_client() as cache:
            result = bool(await cache.time())

        return result

    def get_cache_client(self, *, alone=False):

        client = None

        if alone:

            client = self._redis_pool.get_client()

        else:

            client = self._cache_client_context.get()

            if client is None:

                client = self._redis_pool.get_client()

                if client:
                    self._cache_client_context.set(client)

        return client

    def share_cache(self, cache, ckey):

        return ShareCache(cache, ckey)

    def event_dispatcher(self, channel_name, channel_count):

        return DistributedEvent(self._redis_pool, channel_name, channel_count)


class CacheClient(aioredis.Redis, AsyncContextManager):
    """Redis客户端对象，使用with进行上下文管理

    将连接委托给客户端对象管理，提高了整体连接的使用率

    """

    def __init__(self, pool, expire, key_prefix):

        super().__init__(None)

        self._pool = pool

        self._expire = expire

        self._key_prefix = key_prefix

    async def _init_conn(self):

        global REDIS_POOL_WATER_LEVEL_WARNING_LINE

        if self._pool_or_conn is None and self._pool:

            if (self._pool.maxsize - self._pool.size + self._pool.freesize) < REDIS_POOL_WATER_LEVEL_WARNING_LINE:
                Utils.log.warning(
                    f'Redis connection pool not enough: '
                    f'{self._pool.freesize}({self._pool.size}/{self._pool.maxsize})'
                )
            else:
                Utils.log.debug(
                    f'Redis connection pool info: '
                    f'{self._pool.freesize}({self._pool.size}/{self._pool.maxsize})'
                )

            self._pool_or_conn = await self._pool.acquire()

    async def _close_conn(self, discard=False):

        if self._pool and self._pool_or_conn:

            connection, self._pool_or_conn = self._pool_or_conn, None

            if discard:
                connection.close()
                await connection.wait_closed()

            self._pool.release(connection)

    async def _context_release(self):

        await self._close_conn()

    async def release(self):

        await self._close_conn()

    async def _safe_execute(self, func, *args, **kwargs):

        global REDIS_ERROR_RETRY_COUNT

        result = None

        async for times in AsyncCirculator(max_times=REDIS_ERROR_RETRY_COUNT):

            try:

                await self._init_conn()

                result = await func(*args, **kwargs)

            except (ReplyError, MaxClientsError, AuthError, ReadOnlyError) as err:

                await self._close_conn(True)

                raise err

            except Exception as err:

                await self._close_conn(True)

                if times < REDIS_ERROR_RETRY_COUNT:
                    Utils.log.exception(err)
                else:
                    raise err

            else:

                break

        return result

    async def execute(self, command, *args, **kwargs):

        return await self._safe_execute(super().execute, command, *args, **kwargs)

    @staticmethod
    def _list_basestring(_list):

        return [Utils.basestring(i) for i in _list]

    @staticmethod
    def _val_encode(val):

        return Utils.pickle_dumps(val)

    @staticmethod
    def _val_decode(val):

        return Utils.pickle_loads(val)

    def _list_decode(self, _list):

        return map(
            lambda i: self._val_decode(i) if i else i,
            _list
        )

    def key(self, key, *args, **kwargs):

        if self._key_prefix:
            key = f'{self._key_prefix}_{key}'

        if not args and not kwargs:
            return key

        sign = Utils.params_sign(*args, **kwargs)

        return f'{key}_{sign}'

    def allocate_lock(self, key, expire=60):

        return MLock(self, key, expire)

    # TRANSACTION COMMANDS

    async def unwatch(self):

        return await self._safe_execute(super().unwatch)

    async def watch(self, key, *keys):

        return await self._safe_execute(super().watch, key, *keys)

    def multi_exec(self):

        return MultiExec(self._pool, aioredis.Redis)

    def pipeline(self):

        return Pipeline(self._pool, aioredis.Redis)

    # PUB/SUB COMMANDS

    async def subscribe(self, channel, *channels):

        return await self._safe_execute(super().subscribe, channel, *channels)

    async def unsubscribe(self, channel, *channels):

        return await self._safe_execute(super().unsubscribe, channel, *channels)

    async def psubscribe(self, pattern, *patterns):

        return await self._safe_execute(super().psubscribe, pattern, *patterns)

    async def punsubscribe(self, pattern, *patterns):

        return await self._safe_execute(super().punsubscribe, pattern, *patterns)

    @property
    def channels(self):

        if self._pool_or_conn is None:
            return None

        return self._pool_or_conn.pubsub_channels

    @property
    def patterns(self):

        if self._pool_or_conn is None:
            return None

        return self._pool_or_conn.pubsub_patterns

    @property
    def in_pubsub(self):

        if self._pool_or_conn is None:
            return None

        return self._pool_or_conn.in_pubsub

    # GENERIC COMMANDS

    async def delete(self, *keys):

        _keys = []

        for key in keys:

            if key.find(r'*') < 0:
                _keys.append(key)
            else:
                _keys.extend((await self.keys(key)))

        result = (await super().delete(*_keys)) if len(_keys) > 0 else 0

        return result

    def _delete(self, key, *keys):

        return super().delete(key, *keys)

    async def keys(self, pattern):

        result = await super().keys(pattern)

        if result is not None:
            result = [Utils.basestring(key) for key in result]

        return result

    def _keys(self, pattern, *, encoding=_NOTSET):

        return super().keys(pattern, encoding=encoding)

    async def randomkey(self):

        result = await super().randomkey()

        if result is not None:
            result = Utils.basestring(result)

        return result

    def _randomkey(self, *, encoding=_NOTSET):

        return super().randomkey(encoding=encoding)

    async def scan(self, cursor=0, match=None, count=None):

        result = await super().scan(cursor, match, count)

        if result is not None:
            result = (result[0], self._list_basestring(result[1]))

        return result

    def _scan(self, cursor=0, match=None, count=None):

        return super().scan(cursor, match, count)

    # STRING COMMANDS

    async def get(self, key):

        result = await super().get(key)

        if result is not None:
            result = self._val_decode(result)

        return result

    def _get(self, key, *, encoding=_NOTSET):

        return super().get(key, encoding=encoding)

    async def getset(self, key, value):

        _value = self._val_encode(value)

        result = await super().getset(key, _value)

        if result is not None:
            result = self._val_decode(result)

        return result

    def _getset(self, key, value, *, encoding=_NOTSET):

        return super().getset(key, value, encoding=encoding)

    async def mget(self, key, *keys):

        result = await super().mget(key, *keys)

        if result is not None:
            result = list(self._list_decode(result))

        return result

    def _mget(self, key, *keys, encoding=_NOTSET):

        return super().mget(key, *keys, encoding=encoding)

    async def set(self, key, value, expire=0):

        _value = self._val_encode(value)
        _expire = expire if expire > 0 else self._expire

        result = await super().set(key, _value, expire=_expire)

        return result

    def _set(self, key, value, *, expire=0, pexpire=0, exist=None):

        return super().set(key, value, expire=expire, pexpire=pexpire, exist=exist)

    async def mset(self, key, value, *pairs):

        _value = self._val_encode(value)
        _pairs = [item if index % 2 == 0 else self._val_encode(item) for index, item in enumerate(pairs)]

        result = await super().mset(key, _value, *_pairs)

        return result

    def _mset(self, key, value, *pairs):

        return super().mset(key, value, *pairs)

    async def msetnx(self, key, value, *pairs):

        _value = self._val_encode(value)
        _pairs = [item if index % 2 == 0 else self._val_encode(item) for index, item in enumerate(pairs)]

        result = await super().msetnx(key, _value, *_pairs)

        return result

    def _msetnx(self, key, value, *pairs):

        return super().msetnx(key, value, *pairs)

    async def psetex(self, key, milliseconds, value):

        _value = self._val_encode(value)

        result = await super().psetex(key, milliseconds, _value)

        return result

    def _psetex(self, key, milliseconds, value):

        return super().psetex(key, milliseconds, value)

    async def setex(self, key, seconds, value):

        _value = self._val_encode(value)

        result = await super().setex(key, seconds, _value)

        return result

    def _setex(self, key, seconds, value):

        return super().setex(key, seconds, value)

    async def setnx(self, key, value):

        _value = self._val_encode(value)

        result = await super().setnx(key, _value)

        return result

    def _setnx(self, key, value):

        return super().setnx(key, value)

    # SET COMMANDS

    async def sdiff(self, key, *keys):

        result = await super().sdiff(key, *keys)

        if result is not None:
            result = self._list_basestring(result)

        return result

    def _sdiff(self, key, *keys):

        return super().sdiff(key, *keys)

    async def sinter(self, key, *keys):

        result = await super().sinter(key, *keys)

        if result is not None:
            result = self._list_basestring(result)

        return result

    def _sinter(self, key, *keys):

        return super().sinter(key, *keys)

    async def smembers(self, key):

        result = await super().smembers(key)

        if result is not None:
            result = self._list_basestring(result)

        return result

    def _smembers(self, key, *, encoding=_NOTSET):

        return super().smembers(key, encoding=encoding)

    async def spop(self, key):

        result = await super().spop(key)

        if result is not None:
            result = Utils.basestring(result)

        return result

    def _spop(self, key, *, encoding=_NOTSET):

        return super().spop(key, encoding=encoding)

    async def srandmember(self, key, count=1):

        result = await super().srandmember(key, count)

        if result is not None:
            result = self._list_basestring(result)

        return result

    def _srandmember(self, key, count=None):

        return super().srandmember(key, count)

    async def sunion(self, key, *keys):

        result = await super().sunion(key, *keys)

        if result is not None:
            result = self._list_basestring(result)

        return result

    def _sunion(self, key, *keys):

        return super().sunion(key, *keys)

    async def sscan(self, key, cursor=0, match=None, count=None):

        result = await super().sscan(key, cursor, match, count)

        if result is not None:
            result = (result[0], self._list_basestring(result[1]))

        return result

    def _sscan(self, key, cursor=0, match=None, count=None):

        return super().sscan(key, cursor, match, count)

    # HASH COMMANDS

    async def hget(self, key, field):

        result = await super().hget(key, field)

        if result is not None:
            result = self._val_decode(result)

        return result

    def _hget(self, key, field, *, encoding=_NOTSET):

        return super().hget(key, field, encoding=encoding)

    async def hgetall(self, key):

        result = await super().hgetall(key)

        if result is not None:
            result = {Utils.basestring(key): self._val_decode(val) for key, val in result.items()}

        return result

    def _hgetall(self, key, *, encoding=_NOTSET):

        return super().hgetall(key, encoding=encoding)

    async def hkeys(self, key):

        result = await super().hkeys(key)

        if result is not None:
            result = [Utils.basestring(key) for key in result]

        return result

    def _hkeys(self, key, *, encoding=_NOTSET):

        return super().hkeys(key, encoding=encoding)

    async def hmget(self, key, field, *fields):

        result = await super().hmget(key, field, *fields)

        if result is not None:
            result = list(self._list_decode(result))

        return result

    def _hmget(self, key, field, *fields, encoding=_NOTSET):

        return super().hmget(key, field, *fields, encoding=encoding)

    async def hmset(self, key, field, value, *pairs):

        _value = self._val_encode(value)
        _pairs = [item if index % 2 == 0 else self._val_encode(item) for index, item in enumerate(pairs)]

        result = await super().hmset(key, field, _value, *_pairs)

        return result

    def _hmset(self, key, field, value, *pairs):

        return super().hmset(key, field, value, *pairs)

    async def hmset_dict(self, key, *args, **kwargs):

        for _arg in args:
            if isinstance(_arg, dict):
                kwargs.update({key: self._val_encode(val) for key, val in _arg.items()})

        return await super().hmset_dict(key, kwargs)

    def _hmset_dict(self, key, *args, **kwargs):

        return super().hmset_dict(key, *args, **kwargs)

    async def hset(self, key, field, value):

        _value = self._val_encode(value)

        result = await super().hset(key, field, _value)

        return result

    def _hset(self, key, field, value):

        return super().hset(key, field, value)

    async def hsetnx(self, key, field, value):

        _value = self._val_encode(value)

        result = await super().hsetnx(key, field, _value)

        return result

    def _hsetnx(self, key, field, value):

        return super().hsetnx(key, field, value)

    async def hvals(self, key):

        result = await super().hvals(key)

        if result is not None:
            result = list(self._list_decode(result))

        return result

    def _hvals(self, key, *, encoding=_NOTSET):

        return super().hvals(key, encoding=encoding)

    async def hscan(self, key, cursor=0, match=None, count=None):

        result = await super().hscan(key, cursor, match, count)

        if result is not None:
            result = (result[0], [(Utils.basestring(key), self._val_decode(val),) for key, val in result[1]])

        return result

    def _hscan(self, key, cursor=0, match=None, count=None):

        return super().hscan(key, cursor, match, count)

    # LIST COMMANDS

    async def blpop(self, key, *, timeout=0):

        result = await super().blpop(key, timeout=timeout)

        if result is not None:
            result = self._val_decode(result[1])

        return result

    def _blpop(self, key, *keys, timeout=0, encoding=_NOTSET):

        return super().blpop(key, *keys, timeout=timeout, encoding=encoding)

    async def brpop(self, key, *, timeout=0):

        result = await super().brpop(key, timeout=timeout)

        if result is not None:
            result = self._val_decode(result[1])

        return result

    def _brpop(self, key, *keys, timeout=0, encoding=_NOTSET):

        return super().brpop(key, *keys, timeout=timeout, encoding=encoding)

    async def brpoplpush(self, sourcekey, destkey, timeout=0):

        result = await super().brpoplpush(sourcekey, destkey, timeout)

        if result is not None:
            result = self._val_decode(result)

        return result

    def _brpoplpush(self, sourcekey, destkey, timeout=0, encoding=_NOTSET):

        return super().brpoplpush(sourcekey, destkey, timeout, encoding)

    async def lindex(self, key, index):

        result = await super().lindex(key, index)

        if result is not None:
            result = self._val_decode(result)

        return result

    def _lindex(self, key, index, *, encoding=_NOTSET):

        return super().lindex(key, index, encoding=encoding)

    async def linsert(self, key, pivot, value, before=False):

        _pivot = self._val_encode(pivot)
        _value = self._val_encode(value)

        result = await super().linsert(key, _pivot, _value, before=before)

        return result

    def _linsert(self, key, pivot, value, before=False):

        return super().linsert(key, pivot, value, before=before)

    async def lpop(self, key):

        result = await super().lpop(key)

        if result is not None:
            result = self._val_decode(result)

        return result

    def _lpop(self, key, *, encoding=_NOTSET):

        return super().lpop(key, encoding=encoding)

    async def lpush(self, key, value, *values):

        _value = self._val_encode(value)
        _values = [self._val_encode(val) for val in values]

        result = await super().lpush(key, _value, *_values)

        return result

    def _lpush(self, key, value, *values):

        return super().lpush(key, value, *values)

    async def lpushx(self, key, value):

        _value = self._val_encode(value)

        result = await super().lpushx(key, _value)

        return result

    def _lpushx(self, key, value):

        return super().lpushx(key, value)

    async def lrange(self, key, start, stop):

        result = await super().lrange(key, start, stop)

        if result is not None:
            result = list(self._list_decode(result))

        return result

    def _lrange(self, key, start, stop, *, encoding=_NOTSET):

        return super().lrange(key, start, stop, encoding=encoding)

    async def lrem(self, key, count, value):

        _value = self._val_encode(value)

        result = await super().lrem(key, count, _value)

        return result

    def _lrem(self, key, count, value):

        return super().lrem(key, count, value)

    async def lset(self, key, index, value):

        _value = self._val_encode(value)

        result = await super().lset(key, index, _value)

        return result

    def _lset(self, key, index, value):

        return super().lset(key, index, value)

    async def rpop(self, key):

        result = await super().rpop(key)

        if result is not None:
            result = self._val_decode(result)

        return result

    def _rpop(self, key, *, encoding=_NOTSET):

        return super().rpop(self, key, encoding=encoding)

    async def rpoplpush(self, sourcekey, destkey):

        result = await super().rpoplpush(sourcekey, destkey)

        if result is not None:
            result = self._val_decode(result)

        return result

    def _rpoplpush(self, sourcekey, destkey, *, encoding=_NOTSET):

        return super().rpoplpush(sourcekey, destkey, encoding=encoding)

    async def rpush(self, key, value, *values):

        _value = self._val_encode(value)
        _values = [self._val_encode(val) for val in values]

        result = await super().rpush(key, _value, *_values)

        return result

    def _rpush(self, key, value, *values):

        return super().rpush(key, value, *values)

    async def rpushx(self, key, value):

        _value = self._val_encode(value)

        result = await super().rpushx(key, _value)

        return result

    def _rpushx(self, key, value):

        return super().rpushx(key, value)


class MLock(AsyncContextManager):
    """基于Redis实现的分布式锁，使用with进行上下文管理
    """

    _renew_script = '''
if redis.call("get",KEYS[1]) == ARGV[1] and redis.call("ttl",KEYS[1]) > 0 then
    return redis.call("expire",KEYS[1],ARGV[2])
else
    return 0
end
'''

    _unlock_script = '''
if redis.call("get",KEYS[1]) == ARGV[1] then
    return redis.call("del",KEYS[1])
else
    return 0
end
'''

    def __init__(self, cache, key, expire):

        self._cache = cache
        self._expire = expire

        self._lock_tag = f'process_lock_{key}'
        self._lock_val = Utils.uuid1().encode()

        self._locked = False

    @property
    def locked(self):

        return self._locked

    async def _context_release(self):

        await self.release()

    async def exists(self):

        if await self._cache.exists(self._lock_tag):
            return True
        else:
            return False

    async def acquire(self, timeout=0):

        if self._locked:

            await self.renew()

        else:

            params = {
                r'key': self._lock_tag,
                r'value': self._lock_val,
                r'expire': self._expire,
                r'exist': StringCommandsMixin.SET_IF_NOT_EXIST,
            }

            async for _ in AsyncCirculator(timeout):

                if await self._cache._set(**params):
                    self._locked = True

                if self._locked or timeout == 0:
                    break

        return self._locked

    async def wait(self, timeout=0):

        async for _ in AsyncCirculator(timeout):

            if not await self.exists():
                return True
        else:

            return False

    async def renew(self):

        if self._locked:

            if await self._cache.eval(self._renew_script, [self._lock_tag], [self._lock_val, self._expire]):
                self._locked = True
            else:
                self._locked = False

        return self._locked

    async def release(self):

        if self._locked:
            await self._cache.eval(self._unlock_script, [self._lock_tag], [self._lock_val])
            self._locked = False


class ShareCache(AsyncContextManager):
    """共享缓存，使用with进行上下文管理

    基于分布式锁实现的一个缓存共享逻辑，保证在分布式环境下，同一时刻业务逻辑只执行一次，其运行结果会通过缓存被共享

    """

    def __init__(self, cache, ckey):

        self._cache = cache
        self._ckey = ckey

        self._locker = None
        self._locked = False

        self.result = None

    async def _context_release(self):

        await self.release()

    async def get(self):

        result = await self._cache.get(self._ckey)

        if result is None:

            self._locker = self._cache.allocate_lock(self._ckey)
            self._locked = await self._locker.acquire()

            if not self._locked:
                await self._locker.wait()
                result = await self._cache.get(self._ckey)

        return result

    async def set(self, value, expire=0):

        result = await self._cache.set(self._ckey, value, expire)

        return result

    async def release(self):

        if self._locked:

            self._locked = False

            if self._locker:
                await self._locker.release()

        self._cache = self._locker = None



from najapy.frame.fastapi.base import APIRouter
from najapy.scaffold.fastapi_example_project.fastapi_example.abc_base.view import home_router

api_router = APIRouter()

EXTERNAL_PREFIX = '/external'
INTERNAL_PREFIX = '/internal'
CRONTAB_PREFIX = '/crontab'

##################################################
# 根路径
api_router.include_router(home_router, tags=['home'])

##################################################
# crontab路径


##################################################
# external路径

##################################################
# internal路径



from fastapi import FastAPI
from starlette.middleware.cors import CORSMiddleware
from najapy.frame.fastapi.base import create_fastapi

from najapy.scaffold.fastapi_example_project.fastapi_example.activate import ActivateInit
from najapy.scaffold.fastapi_example_project.fastapi_example.conf import ConfigDynamic
from najapy.scaffold.fastapi_example_project.fastapi_example.err_handler import exception_handler
from najapy.scaffold.fastapi_example_project.fastapi_example.router import api_router


def get_application() -> FastAPI:
    fastapi = create_fastapi(
        log_level=ConfigDynamic.LogLevel,
        debug=ConfigDynamic.Debug,
        on_startup=[ActivateInit.activate_initialize],
        on_shutdown=[ActivateInit.activate_release],
        openapi_url=r'/openapi.json',
        docs_url=r'/docs',
        title=r'fastapi_example_project',
    )

    if ConfigDynamic.AllowedHosts:
        fastapi.add_middleware(
            CORSMiddleware,
            allow_origins=ConfigDynamic.AbnormalStatus,
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )

    fastapi.include_router(api_router, prefix='/api/v1/fastapi_example')

    exception_handler(fastapi)

    return fastapi


app = get_application()



from najapy.common.metaclass import Singleton


class Enum:

    def __init__(self, code, message, value=0):
        self.code = code
        self.message = message
        self.value = value

    def __eq__(self, other):
        if isinstance(other, Enum):
            return self.code == other.code
        elif self.code == other:
            return True
        else:
            return False


class BaseEnum(Singleton):
    _attr_list = None
    _attr_dict = None
    _attr_dict_enum = None
    _attr_weight_dict = None

    @classmethod
    def to_list(cls):

        if cls._attr_list:
            return cls._attr_list

        attr_list = []

        for attr in dir(cls):

            if not attr.startswith("__"):
                attr_obj = getattr(cls, attr)
                if isinstance(attr_obj, Enum):
                    attr_list.append(attr_obj)
        attr_list = sorted([attr.code for attr in attr_list])

        cls._attr_list = attr_list
        return attr_list

    @classmethod
    def to_dict(cls):

        if cls._attr_dict:
            return cls._attr_dict

        attr_dict = {}

        for attr in dir(cls):

            if not attr.startswith("__"):
                attr_obj = getattr(cls, attr)
                if isinstance(attr_obj, Enum):
                    attr_dict[attr_obj.code] = attr_obj.message

        cls._attr_dict = attr_dict
        return attr_dict

    @classmethod
    def to_dict_enum(cls):

        if cls._attr_dict_enum:
            return cls._attr_dict_enum

        attr_dict_enum = {}

        for attr in dir(cls):

            if not attr.startswith("__"):
                attr_obj = getattr(cls, attr)
                if isinstance(attr_obj, Enum):
                    attr_dict_enum[attr_obj.code] = attr_obj

        cls._attr_dict_enum = attr_dict_enum
        return attr_dict_enum

    @classmethod
    def to_weight(cls):

        if cls._attr_weight_dict:
            return cls._attr_weight_dict

        _attr_weight_dict = {}

        for attr in dir(cls):

            if not attr.startswith("__"):
                attr_obj = getattr(cls, attr)
                if isinstance(attr_obj, Enum):
                    _attr_weight_dict[attr_obj.code] = attr_obj.value

        cls._attr_weight_dict = _attr_weight_dict
        return _attr_weight_dict

    def __iter__(self):
        return iter(self.to_list())

    def __repr__(self):
        return '%s' % self.to_list()



from starlette.responses import JSONResponse

from najapy.common.struct import Const, Result


class RespCode():
    Success = 0  # 成功
    Unknown = -1  # 未知


RespMessage = Const()
RespMessage[RespCode.Success] = r"成功"
RespMessage[RespCode.Unknown] = r"未知"


class Response(JSONResponse):

    def render(self, content):
        return super().render(Result(code=RespCode.Success, data=content, msg=RespMessage[RespCode.Success]))


class ErrorResponse(Exception, JSONResponse):

    def __init__(self, error_code, content=None, msg=None, status_code=200, details=None, **kwargs):
        self._error_code = error_code
        self._details = details
        self._msg = msg

        Exception.__init__(self)
        JSONResponse.__init__(self, content, status_code, **kwargs)

    def render(self, content):
        if not self._msg:
            self._msg = RespMessage[self._error_code]

        return super().render(
            Result(code=self._error_code, data=content, msg=self._msg, details=self._details)
        )



"""任务工具集"""
import pytz
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.triggers.cron import CronTrigger

from najapy.common.async_base import Utils

TIMEZONE = pytz.timezone(r'Asia/Shanghai')


class TaskInterface:
    """Task接口定义
    """

    def start(self):
        raise NotImplementedError()

    def stop(self):
        raise NotImplementedError()

    def is_running(self):
        raise NotImplementedError()


class TaskAbstract(TaskInterface):
    """任务基类
    """

    def __init__(self, scheduler=None):

        global TIMEZONE

        self._scheduler = AsyncIOScheduler(
            job_defaults={
                r'coalesce': False,
                r'max_instances': 1,
                r'misfire_grace_time': 10
            },
            timezone=TIMEZONE
        ) if scheduler is None else scheduler

    @property
    def scheduler(self):

        return self._scheduler

    @staticmethod
    def _func_wrapper(func, *args, **kwargs):

        if Utils.is_coroutine_function(func):

            if args or kwargs:
                return Utils.func_partial(func, *args, **kwargs)
            else:
                return func

        async def _wrapper():
            return func(*args, **kwargs)

        return _wrapper

    def is_running(self):

        return self._scheduler.running

    def start(self):

        return self._scheduler.start()

    def stop(self):

        return self._scheduler.shutdown()

    def add_job(self):

        raise NotImplementedError()

    def remove_job(self, job_id):

        return self._scheduler.remove_job(job_id)

    def remove_all_jobs(self):

        return self._scheduler.remove_all_jobs()


class IntervalTask(TaskAbstract):
    """间隔任务类
    """

    @classmethod
    def create(cls, interval, func, *args, **kwargs):
        inst = cls()

        inst.add_job(interval, func, *args, **kwargs)

        return inst

    def add_job(self, interval, func, *args, **kwargs):
        return self._scheduler.add_job(
            self._func_wrapper(func, *args, **kwargs),
            r'interval', seconds=interval
        )


class CronTask(TaskAbstract):
    """定时任务类
    """

    @classmethod
    def create(cls, crontab, func, *args, **kwargs):
        inst = cls()

        inst.add_job(crontab, func, *args, **kwargs)

        return inst

    def add_job(self, crontab, func, *args, **kwargs):
        return self._scheduler.add_job(
            self._func_wrapper(func, *args, **kwargs),
            CronTrigger.from_crontab(crontab, TIMEZONE)
        )



import copy
from datetime import datetime
from io import BytesIO
import xlrd
from xlwt import Workbook, XFStyle, Borders, Pattern

from najapy.common.async_base import Utils


class ExcelRD():
    """excel读取类"""

    def __init__(self, filename=None, file_contents=None, sheet_name=None):
        self.filename = filename
        self.file_contents = file_contents

        if self.filename:
            self.data = xlrd.open_workbook(filename=filename)
        else:
            self.data = xlrd.open_workbook(file_contents=file_contents)

        if sheet_name:
            self.sheet = self.data.sheet_by_name(sheet_name)
        else:
            self.sheet = self.data.sheet_by_index(0)

        self.header_infos = self.sheet.row_values(0)  # 表头信息

        self.row_nums = self.sheet.nrows  # 全部行数
        self.col_nums = self.sheet.ncols  # 全部列数

        self._field_dict = None

    @property
    def field_dict(self):
        """
        field_dict: 表头信息和代码的映射关系字典 {"姓名": "name", "考勤组": "section", "UserId": "user_id"}
        """
        return self._field_dict

    @field_dict.setter
    def field_dict(self, value):
        """
        field_dict:  {"name": "姓名", "age": "年龄"}
        """
        self._field_dict = value

    def _get_fields_col_relation_by_field_dict(self):
        """
        根据表头字段，获取字段和所在列的对应关系
        :return: {"name":1, "age": 3}  ==> {字段：所在列}
        """
        row_data = self.sheet.row_values(0)  # 获取表头信息

        new_row_data = []
        for _row_data in row_data:
            if isinstance(_row_data, float):
                _row_data = str(int(_row_data))

            new_row_data.append(_row_data)

        relation = {}
        for field, field_name in self._field_dict.items():
            if not isinstance(field, str):
                continue

            for index, val in enumerate(new_row_data):

                if field_name.strip() == str(val).strip():
                    relation[field] = index

        return relation

    def _get_values_by_fields(self, fields_relation, start=1):
        """
        :param fields_relation: {"name":1, "age": 3}
        :return: [{"name":"李四", "age": "12"}, {"name":"张三", "age": "16"}]
        """

        rows = []
        for index in range(start, self.row_nums):
            row_data = self.sheet.row_values(index)
            row = {}
            fields = fields_relation.keys()

            for field in fields:
                val_index = fields_relation[field]
                row[field] = str(row_data[val_index])

            rows.append(row)

        return rows

    def get_excel_data(self, start=1):
        """
        根据excel表头和脚本的映射关系获取excel数据
        """
        result = None

        relation = self._get_fields_col_relation_by_field_dict()

        if len(relation) != len(self._field_dict):
            shorted_field = []

            for field, field_str in self._field_dict.items():
                if field not in relation:
                    shorted_field.append(field_str)

            Utils.log.error(f"excel_parsing_data rows error, relation:{relation},shorted_field:{shorted_field}")
            return result

        try:
            rows = self._get_values_by_fields(relation, start)
            result = rows
        except Exception as err:
            Utils.log.error(f"excel_parsing_data rows error:{err},relation:{relation}")

        return result

    def get_all_datas(self):
        """
        获取表格内的所有数据并获取单元格行号和列号
        """
        datas = []

        for i in range(1, self.row_nums):

            sheet_dict = {}

            for j in range(self.col_nums):
                # 获取该单元格所在列的表头名称
                header_name = self.header_infos[j]
                # 获取单元格数据
                cell_data = self.sheet.cell_value(i, j)

                # key值为field_dict中表头名称对应的值或者所在列号
                data_key = str(self._field_dict.get(header_name, j))
                sheet_dict[data_key] = {
                    "row_num": i,
                    "col_num": j,
                    "cell_data": cell_data
                }

            datas.append(sheet_dict)

        return datas


class ExcelWT(Workbook):
    """Excel生成工具
    """

    def __init__(self, name, encoding=r'utf-8', style_compression=0):

        super().__init__(encoding, style_compression)

        self._book_name = name
        self._current_sheet = None

        self._default_style = XFStyle()
        self._default_style.borders.left = Borders.THIN
        self._default_style.borders.right = Borders.THIN
        self._default_style.borders.top = Borders.THIN
        self._default_style.borders.bottom = Borders.THIN
        self._default_style.pattern.pattern = Pattern.SOLID_PATTERN
        self._default_style.pattern.pattern_fore_colour = 0x01

        self._default_title_style = copy.deepcopy(self._default_style)
        self._default_title_style.font.bold = True
        self._default_title_style.pattern.pattern_fore_colour = 0x16

    def create_sheet(self, name, titles=None):

        sheet = self._current_sheet = self.add_sheet(name)
        style = self._default_title_style

        if titles:
            for index, title in enumerate(titles):
                sheet.write(0, index, title, style)
                sheet.col(index).width = 0x1200

    def add_sheet_row(self, *args):

        sheet = self._current_sheet
        style = self._default_style

        nrow = len(sheet.rows)

        for index, value in enumerate(args):
            sheet.write(nrow, index, value, style)

    def get_file(self):

        result = b''

        with BytesIO() as stream:
            self.save(stream)

            result = stream.getvalue()

        return result

    def write_request(self, request):

        filename = f"{self._book_name}.{datetime.today().strftime('%y%m%d.%H%M%S')}.xls"

        request.set_header(r'Content-Type', r'application/vnd.ms-excel')
        request.set_header(r'Content-Disposition', f'attachment;filename={filename}')

        return request.finish(self.get_file())



from fastapi import APIRouter
from starlette import status

from najapy.scaffold.fastapi_example_project.fastapi_example.abc_base.service.base_service import DataSource
from najapy.scaffold.fastapi_example_project.fastapi_example.utils.response import Response, ErrorResponse, RespCode

home_router = APIRouter()


@home_router.get(r'/')
async def default():
    if DataSource().online:
        return Response(status_code=status.HTTP_200_OK)
    else:
        return ErrorResponse(error_code=RespCode.Unknown, status_code=status.HTTP_500_INTERNAL_SERVER_ERROR)


@home_router.head(r'/health')
async def default():
    ok = await DataSource().health()
    if not ok:
        return ErrorResponse(error_code=RespCode.Unknown, status_code=status.HTTP_500_INTERNAL_SERVER_ERROR)
    else:
        return Response(status_code=status.HTTP_200_OK)



from distutils.core import setup
from setuptools import find_packages
from najapy import __version__

with open("README.md", "r", encoding='utf-8') as f:
    long_description = f.read()

setup(name=r'najapy',
      version=__version__,
      license=r'Apache License Version 2.0',
      author=r'lanqiao',
      author_email=r'',
      description=r'Async Suite For Python',
      long_description=long_description,
      long_description_content_type=r'text/markdown',
      url=r'https://github.com/lanqiao-dev/najapy.git',
      packages=find_packages(),
      package_data={r'najapy': [r'static/*.*', r'scaffold/fastapi_example_project/fastapi_example/conf/*.*']},
      python_requires=r'>= 3.8',
      platforms=[r"all"],
      install_requires=[
          r'APScheduler==3.7.0',
          r'PyJWT==2.0.1',
          r'SQLAlchemy==1.3.23',
          r'WTForms==2.3.3',
          r'aiohttp==3.7.4',
          r'aiomysql==0.0.21',
          r'aioredis==1.3.1',
          r'async-timeout==3.0.1',
          r'cachetools==4.2.1',
          r'cryptography==3.4.6',
          r'fastapi==0.66.1',
          r'gunicorn==20.1.0',
          r'hiredis==1.1.0',
          r'httptools==0.1.1',
          r'loguru==0.5.3',
          r'motor==2.3.1',
          r'ujson==4.0.2',
          r'pytz==2021.1',
          r'PyMySQL==0.9.3',
          r'pymongo==3.12.0',
          r'uvicorn==0.14.0',
          r'xmltodict==0.12.0',
          r'psutil==5.8.0',
          r'xlwt==1.3.0',
          r'xlrd==2.0.1',
          r'xmltodict==0.12.0'
      ],
      classifiers=[
          r'Programming Language :: Python :: 3.8',
          r'License :: OSI Approved :: Apache Software License',
          r'Operating System :: POSIX :: Linux',
      ],
      )



import os
import ssl
from contextlib import contextmanager

import aiohttp

from najapy.common.async_base import AsyncCirculatorForSecond, Utils

DEFAULT_TIMEOUT = aiohttp.client.ClientTimeout(total=60, connect=10, sock_read=60, sock_connect=10)
DOWNLOAD_TIMEOUT = aiohttp.client.ClientTimeout(total=600, connect=10, sock_read=600, sock_connect=10)

CACERT_FILE = os.path.join(
    os.path.split(os.path.abspath(__file__))[0],
    r'../static/cacert.pem'
)


class _AsyncCirculator(AsyncCirculatorForSecond):

    async def _sleep(self):
        await Utils.sleep(self._current)


def _json_decoder(val, **kwargs):
    try:
        return Utils.json_decode(val, **kwargs)
    except Exception as err:
        Utils.log.warning(f'http client json decode error: {err} => {val}')


@contextmanager
def _catch_error():
    """异常捕获

    通过with语句捕获异常，代码更清晰，还可以使用Ignore异常安全的跳出with代码块

    """

    try:

        yield

    except aiohttp.ClientResponseError as err:

        if err.status < 500:
            Utils.log.warning(err)
        else:
            Utils.log.error(err)

    except Exception as err:

        Utils.log.error(err)


class Result(dict):

    def __init__(self, status, headers, body):
        super().__init__(status=status, headers=headers, body=body)

    def __bool__(self):
        return (self.status >= 200) and (self.status <= 299)

    @property
    def status(self):
        return self.get(r'status')

    @property
    def headers(self):
        return self.get(r'headers')

    @property
    def body(self):
        return self.get(r'body')


class _HTTPClient:
    """HTTP客户端基类
    """

    def __init__(self, retry_count=5, timeout=None, raise_for_status=True, **kwargs):

        global DEFAULT_TIMEOUT

        self._ssl_context = self._create_ssl_context()

        self._retry_count = retry_count

        self._session_config = kwargs
        self._session_config[r'timeout'] = timeout if timeout is not None else DEFAULT_TIMEOUT
        self._session_config.setdefault(r'raise_for_status', raise_for_status)

    async def _handle_response(self, response):

        return await response.read()

    def _create_ssl_context(self):

        global CACERT_FILE

        return ssl.create_default_context(cafile=CACERT_FILE)

    async def send_request(
            self, method, url, data=None, params=None, cookies=None, headers=None, **settings
    ):

        response = None

        if headers is None:
            headers = {}

        settings[r'data'] = data
        settings[r'params'] = params
        settings[r'cookies'] = cookies
        settings[r'headers'] = headers

        Utils.log.debug(
            r'{0} {1} => {2}'.format(
                method,
                url,
                str({key: val for key, val in settings.items() if isinstance(val, (str, list, dict))})
            )
        )

        settings.setdefault(r'ssl', self._ssl_context)

        async for times in _AsyncCirculator(max_times=self._retry_count):

            try:

                async with aiohttp.ClientSession(**self._session_config) as _session:

                    async with _session.request(method, url, **settings) as _response:
                        response = Result(
                            _response.status,
                            dict(_response.headers),
                            await self._handle_response(_response)
                        )

            except aiohttp.ClientResponseError as err:

                # 重新尝试的话，会记录异常，否则会继续抛出异常

                if err.status < 500:
                    raise err
                elif times >= self._retry_count:
                    raise err
                else:
                    Utils.log.warning(err)
                    continue

            except aiohttp.ClientError as err:

                if times >= self._retry_count:
                    raise err
                else:
                    Utils.log.warning(err)
                    continue

            except Exception as err:

                raise err

            else:

                Utils.log.info(f'{method} {url} => status:{response.status}')
                break

            finally:

                if times > 1:
                    Utils.log.warning(f'{method} {url} => retry:{times}')

        return response


class _HTTPJsonMixin:
    """Json模式混入类
    """

    async def _handle_response(self, response):
        return await response.json(encoding=r'utf-8', loads=_json_decoder, content_type=None)


class HTTPClient(_HTTPClient):
    """HTTP客户端，普通模式
    """

    async def get(self, url, params=None, *, cookies=None, headers=None):
        result = None

        with _catch_error():
            resp = await self.send_request(aiohttp.hdrs.METH_GET, url, None, params, cookies=cookies, headers=headers)

            result = resp.body

        return result

    async def options(self, url, params=None, *, cookies=None, headers=None):
        result = None

        with _catch_error():
            resp = await self.send_request(aiohttp.hdrs.METH_OPTIONS, url, None, params, cookies=cookies,
                                           headers=headers)

            result = resp.headers

        return result

    async def head(self, url, params=None, *, cookies=None, headers=None):
        result = None

        with _catch_error():
            resp = await self.send_request(aiohttp.hdrs.METH_HEAD, url, None, params, cookies=cookies, headers=headers)

            result = resp.headers

        return result

    async def post(self, url, data=None, params=None, *, cookies=None, headers=None):
        result = None

        with _catch_error():
            resp = await self.send_request(aiohttp.hdrs.METH_POST, url, data, params, cookies=cookies, headers=headers)

            result = resp.body

        return result

    async def put(self, url, data=None, params=None, *, cookies=None, headers=None):
        result = None

        with _catch_error():
            resp = await self.send_request(aiohttp.hdrs.METH_PUT, url, data, params, cookies=cookies, headers=headers)

            result = resp.body

        return result

    async def patch(self, url, data=None, params=None, *, cookies=None, headers=None):
        result = None

        with _catch_error():
            resp = await self.send_request(aiohttp.hdrs.METH_PATCH, url, data, params, cookies=cookies, headers=headers)

            result = resp.body

        return result

    async def delete(self, url, params=None, *, cookies=None, headers=None):
        result = None

        with _catch_error():
            resp = await self.send_request(aiohttp.hdrs.METH_DELETE, url, None, params, cookies=cookies,
                                           headers=headers)

            result = resp.body

        return result


class HTTPJsonClient(_HTTPJsonMixin, HTTPClient):
    """HTTP客户端，Json模式
    """
    pass


class HTTPClientPool(HTTPClient):
    """HTTP带连接池客户端，普通模式
    """

    def __init__(self,
                 retry_count=5, use_dns_cache=True, ttl_dns_cache=10,
                 limit=100, limit_per_host=0, timeout=None, raise_for_status=True,
                 **kwargs
                 ):
        super().__init__(retry_count, timeout, raise_for_status, **kwargs)

        self._tcp_connector = aiohttp.TCPConnector(
            use_dns_cache=use_dns_cache,
            ttl_dns_cache=ttl_dns_cache,
            ssl=self._ssl_context,
            limit=limit,
            limit_per_host=limit_per_host,
        )

        self._session_config[r'connector'] = self._tcp_connector
        self._session_config[r'connector_owner'] = False

    async def close(self):
        if not self._tcp_connector.closed:
            await self._tcp_connector.close()


class HTTPJsonClientPool(_HTTPJsonMixin, HTTPClientPool):
    """HTTP带连接池客户端，Json模式
    """
    pass



from najapy.scaffold.fastapi_example_project.fastapi_example.abc_base.service.base_service import DataSource


class ActivateInit:

    @staticmethod
    async def activate_initialize():
        await DataSource.initialize()

    @staticmethod
    async def activate_release():
        await DataSource().release()



from najapy.database.mongo import MongoDelegate
from najapy.cache.base import ShareFuture, FuncCache
from najapy.cache.redis import RedisDelegate
from najapy.common.async_base import MultiTasks, Utils
from najapy.common.base import catch_error, Ignore
from najapy.common.metaclass import Singleton
from najapy.common.process import HeartbeatChecker
from najapy.database.mysql import MySQLDelegate

from najapy.scaffold.fastapi_example_project.fastapi_example.conf import ConfigStatic, ConfigDynamic


class DataSource(Singleton, RedisDelegate, MySQLDelegate, MongoDelegate):

    def __init__(self):
        MySQLDelegate.__init__(self)
        RedisDelegate.__init__(self)
        MongoDelegate.__init__(
            self,
            host=ConfigStatic.MongoServer,
            username=ConfigStatic.MongoUser,
            password=ConfigStatic.MongoPasswd,
            min_pool_size=ConfigStatic.MongoMinConn,
            max_pool_size=ConfigStatic.MongoMaxConn,
            max_idle_time=3600
        )

        self._heartbeat = HeartbeatChecker()

        self._event_bus = None

    @classmethod
    async def initialize(cls):

        inst = cls()

        await inst.async_init_redis(
            ConfigStatic.RedisHost, ConfigStatic.RedisPasswd,
            minsize=ConfigStatic.RedisMinConn, maxsize=ConfigStatic.RedisMaxConn,
            db=ConfigStatic.RedisBase, expire=ConfigStatic.RedisExpire,
            key_prefix=ConfigStatic.RedisKeyPrefix
        )

        # 初始化mysql连接
        await inst.async_init_mysql_rw(
            ConfigStatic.MySqlMasterServer[0], ConfigStatic.MySqlMasterServer[1], ConfigStatic.MySqlName,
            ConfigStatic.MySqlUser, ConfigStatic.MySqlPasswd,
            minsize=ConfigStatic.MySqlMasterMinConn, maxsize=ConfigStatic.MySqlMasterMaxConn,
            echo=ConfigDynamic.Debug, pool_recycle=21600, conn_life=43200
        )

        if ConfigStatic.MySqlSlaveServer:
            await inst.async_init_mysql_ro(
                ConfigStatic.MySqlSlaveServer[0], ConfigStatic.MySqlSlaveServer[1], ConfigStatic.MySqlName,
                ConfigStatic.MySqlUser, ConfigStatic.MySqlPasswd,
                minsize=ConfigStatic.MySqlSlaveMinConn, maxsize=ConfigStatic.MySqlSlaveMaxConn,
                echo=ConfigDynamic.Debug, readonly=True, pool_recycle=21600, conn_life=43200
            )

        inst._event_bus = inst.event_dispatcher(
            f'{ConfigStatic.RedisKeyPrefix}_client_event',
            ConfigDynamic.ClentEventBusChannels
        )

    async def release(self):

        self._heartbeat.release()

        await self.async_close_redis()
        await self.async_close_mysql()
        await self.close_mongo_pool()

    @property
    def online(self):
        return self._heartbeat.check()

    @ShareFuture()
    @FuncCache(ttl=30)
    async def health(self):
        result = False

        with catch_error():
            tasks = MultiTasks()

            tasks.append(self.cache_health())
            tasks.append(self.mysql_health())
            tasks.append(self.mongo_health())

            await tasks

            result = all(tasks)

            if result:
                self._heartbeat.refresh()

            Utils.log.info(f'Check health {result}')

        return result

    @property
    def event_bus(self):

        return self._event_bus


class ServiceBase(Utils, Singleton):
    """service基类"""

    def __init__(self):
        self._data_source = DataSource()

    @property
    def data_source(self):

        return self._data_source

    def Break(self, data=None, layers=1):

        raise Ignore(data, layers)
